"""This module only contains the Node class.

Author: Thomas Oberbichler
"""

from typing import List
import numpy as np
from nfem.dof import Dof


class Node:
    """Three dimensional Node providing Dofs for displacements.

    Attributes
    ----------
    id : str
        Unique ID.
    ref_x : float
        ref X coordinate.
    ref_y : float
        ref Y coordinate.
    ref_z : float
        ref Z coordinate.
    x : float
        Actual X coordinate.
    y : float
        Actual Y coordinate.
    z : float
        Actual Z coordinate.
    u : float
        Displacement in x direction.
    v : float
        Displacement in y direction.
    w : float
        Displacement in z direction.
    """
    id: str
    _dof_x: Dof
    _dof_y: Dof
    _dof_z: Dof

    def __init__(self, id: str, x: float, y: float, z: float):
        """Create a new node.

        Parameters
        ----------
        id : str
            Unique ID of the node.
        x : float
            Initial X coordinate of the node.
        y : float
            Initial Y coordinate of the node.
        z : float
            Initial Z coordinate of the node.
        """
        self.id = id
        self._dof_x = Dof(id=(id, 'u'), value=x)
        self._dof_y = Dof(id=(id, 'v'), value=y)
        self._dof_z = Dof(id=(id, 'w'), value=z)

    def dof(self, dof_type: str) -> Dof:
        if dof_type == 'u':
            return self._dof_x
        if dof_type == 'v':
            return self._dof_y
        if dof_type == 'w':
            return self._dof_z
        raise AttributeError('Node has no dof of type \'{}\''.format(dof_type))

    @property
    def ref_x(self) -> float:
        """Gets or sets the x coordinate of the node in the undeformed reference configuration."""
        return self._dof_x.ref_value

    @ref_x.setter
    def ref_x(self, value: float):
        self._dof_x.ref_value = value

    @property
    def ref_y(self) -> float:
        """Gets or sets the y coordinate of the node in the undeformed reference configuration."""
        return self._dof_y.ref_value

    @ref_y.setter
    def ref_y(self, value: float):
        self._dof_y.ref_value = value

    @property
    def ref_z(self) -> float:
        """Gets or sets the z coordinate of the node in the undeformed reference configuration."""
        return self._dof_z.ref_value

    @ref_z.setter
    def ref_z(self, value: float):
        self._dof_z.ref_value = value

    @property
    def x(self) -> float:
        return self._dof_x.value

    @x.setter
    def x(self, value: float):
        self._dof_x.value = value

    @property
    def y(self) -> float:
        return self._dof_y.value

    @y.setter
    def y(self, value: float):
        self._dof_y.value = value

    @property
    def z(self) -> float:
        return self._dof_z.value

    @z.setter
    def z(self, value: float):
        self._dof_z.value = value

    @property
    def u(self) -> float:
        return self._dof_x.delta

    @u.setter
    def u(self, value: float):
        self._dof_x.delta = value

    @property
    def v(self) -> float:
        return self._dof_y.delta

    @v.setter
    def v(self, value: float):
        self._dof_y.delta = value

    @property
    def w(self) -> float:
        return self._dof_z.delta

    @w.setter
    def w(self, value: float):
        self._dof_z.delta = value

    @property
    def fx(self) -> float:
        return self._dof_x.external_force

    @fx.setter
    def fx(self, value: float):
        self._dof_x.external_force = value

    @property
    def fy(self) -> float:
        return self._dof_y.external_force

    @fy.setter
    def fy(self, value: float):
        self._dof_y.external_force = value

    @property
    def fz(self) -> float:
        return self._dof_z.external_force

    @fz.setter
    def fz(self, value: float):
        self._dof_z.external_force = value

    @property
    def external_force(self):
        return np.array([self.fx, self.fy, self.fz])

    @external_force.setter
    def external_force(self, value):
        [self.fx, self.fy, self.fz] = value

    @property
    def ref_location(self) -> List[float]:
        """Gets or sets the z coordinate of the node in the undeformed reference configuration."""
        return np.array([self.ref_x, self.ref_y, self.ref_z])

    @ref_location.setter
    def ref_location(self, value):
        self.ref_x, self.ref_y, self.ref_z = value

    @property
    def location(self):
        return np.array([self.x, self.y, self.z])

    @location.setter
    def location(self, value):
        self.x, self.y, self.z = value

    @property
    def displacement(self):
        return np.array([self.u, self.v, self.w])

    @displacement.setter
    def displacement(self, value):
        self.u, self.v, self.w = value

    @property
    def support(self) -> str:
        result = ''
        if self.support_x:
            result += 'x'
        if self.support_y:
            result += 'y'
        if self.support_z:
            result += 'z'
        return result

    @support.setter
    def support(self, value: str):
        self.support_x = 'x' in value
        self.support_y = 'y' in value
        self.support_z = 'z' in value

    @property
    def support_x(self) -> bool:
        return not self._dof_x.is_active

    @support_x.setter
    def support_x(self, value: bool):
        self._dof_x.is_active = not value

    @property
    def support_y(self) -> bool:
        return not self._dof_y.is_active

    @support_y.setter
    def support_y(self, value: bool):
        self._dof_y.is_active = not value

    @property
    def support_z(self) -> bool:
        return not self._dof_z.is_active

    @support_z.setter
    def support_z(self, value: bool):
        self._dof_z.is_active = not value

    def draw(self, item):
        item.set_label_location(self.ref_location, self.location)

        item.add_support(
            location=self.ref_location,
            direction=self.support,
            layer=10,
            color='gray',
        )

        item.add_support(
            location=self.location,
            direction=self.support,
            layer=20,
            color='black',
        )

        item.add_point(
            location=self.ref_location,
            layer=10,
            color='gray',
        )

        item.add_point(
            location=self.location,
            layer=20,
            color='black',
        )

        item.add_result('Location undeformed', self.ref_location.tolist())
        item.add_result('Location', self.location.tolist())
        item.add_result('Displacement', (self.location - self.ref_location).tolist())
        item.add_result('Displacement X', float(self.location[0] - self.ref_location[0]))
        item.add_result('Displacement Y', float(self.location[1] - self.ref_location[1]))
        item.add_result('Displacement Z', float(self.location[2] - self.ref_location[2]))

        if np.linalg.norm(self.external_force) > 1e-8:
            direction = self.external_force
            direction /= np.linalg.norm(direction)

            item.add_arrow(
                location=self.ref_location - direction,
                direction=direction,
                layer=13,
                color='gray',
            )

            item.add_arrow(
                location=self.location - direction,
                direction=direction,
                layer=23,
                color='blue',
            )



"""This module only contains the truss element.

Authors: Thomas Oberbichler, Klaus Sautter
"""

from nfem.node import Node

import numpy as np
import numpy.linalg as la

from typing import Optional


class Truss:
    """FIXME"""

    node_a: Node
    node_b: Node
    youngs_modulus: float
    area: float
    prestress: float
    tensile_strength: Optional[float]
    compressive_strength: Optional[float]

    def __init__(self, id: str, node_a: Node, node_b: Node, youngs_modulus: float, area: float, prestress: float = 0.0, tensile_strength: Optional[float] = None, compressive_strength: Optional[float] = None):
        """FIXME"""

        self.id = id
        self.node_a = node_a
        self.node_b = node_b
        self.youngs_modulus = youngs_modulus
        self.area = area
        self.prestress = prestress
        self.tensile_strength = tensile_strength
        self.compressive_strength = compressive_strength

    @property
    def dofs(self):
        """FIXME"""
        node_a = self.node_a
        node_b = self.node_b

        return [node_a._dof_x, node_a._dof_y, node_a._dof_z, node_b._dof_x, node_b._dof_y, node_b._dof_z]

    @property
    def ref_length(self) -> float:
        """Gets the length of the undeformed truss"""
        a = self.node_a.ref_location
        b = self.node_b.ref_location

        return la.norm(b - a)

    @property
    def length(self) -> float:
        """Gets the length of the deformed truss"""
        a = self.node_a.location
        b = self.node_b.location

        return la.norm(b - a)

    def calculate_elastic_stiffness_matrix(self):
        """FIXME"""
        E = self.youngs_modulus
        A = self.area
        L = self.ref_length

        Dx, Dy, Dz = self.node_b.ref_location - self.node_a.ref_location

        D = np.array([-Dx, -Dy, -Dz, Dx, Dy, Dz])

        return np.outer(E * A / L**3 * D, D)

    def calculate_material_stiffness_matrix(self):
        """FIXME"""
        E = self.youngs_modulus
        A = self.area
        L = self.ref_length

        dx, dy, dz = self.node_b.location - self.node_a.location

        d = np.array([-dx, -dy, -dz, dx, dy, dz])

        return np.outer(E * A / L**3 * d, d)

    def calculate_initial_displacement_stiffness_matrix(self):
        """FIXME"""
        K_m = self.calculate_material_stiffness_matrix()
        K_e = self.calculate_elastic_stiffness_matrix()

        return K_m - K_e

    def calculate_geometric_stiffness_matrix(self, linear: bool = False):
        E = self.youngs_modulus
        A = self.area
        L = self.ref_length
        prestress = self.prestress

        if linear:
            # FIXME this is not a clean solution to solve the LPB issue
            epsilon = self.calculate_linear_strain()
        else:
            epsilon = self.calculate_green_lagrange_strain()

        sigma = E * epsilon + prestress

        q = sigma * A / L

        return np.array([[ q,  0,  0, -q,  0,  0],
                         [ 0,  q,  0,  0, -q,  0],
                         [ 0,  0,  q,  0,  0, -q],
                         [-q,  0,  0,  q,  0,  0],
                         [ 0, -q,  0,  0,  q,  0],
                         [ 0,  0, -q,  0,  0,  q]])

    def calculate_stiffness_matrix(self):
        """FIXME"""

        K_m = self.calculate_material_stiffness_matrix()
        K_g = self.calculate_geometric_stiffness_matrix()

        return K_m + K_g

    def calculate_green_lagrange_strain(self):
        # reference base vector
        A1 = self.node_b.ref_location - self.node_a.ref_location

        # actual base vector
        a1 = self.node_b.location - self.node_a.location

        # green-lagrange strain
        epsilon_GL = (a1 @ a1 - A1 @ A1) / (2 * A1 @ A1)

        return epsilon_GL

    def calculate_linear_strain(self):
        """FIXME"""

        # reference base vector
        A1 = self.node_b.ref_location - self.node_a.ref_location

        # actual base vector
        a1 = self.node_b.location - self.node_a.location

        L = self.ref_length

        # project actual on reference
        projected_l = a1 @ A1 / np.sqrt(A1 @ A1)

        e_lin = (projected_l - L) / L

        return e_lin

    def calculate_internal_forces(self):
        # reference base vector
        A1 = self.node_b.ref_location - self.node_a.ref_location

        # actual base vector
        a1 = self.node_b.location - self.node_a.location

        # green-lagrange strain
        eps = self.calculate_green_lagrange_strain()

        E = self.youngs_modulus
        A = self.area
        L = np.sqrt(A1 @ A1)

        D_pi = (eps * E + self.prestress) * A * L

        D_eps = a1 / (A1 @ A1)

        D_a1 = np.array([[-1,  0,  0,  1,  0,  0],
                         [ 0, -1,  0,  0,  1,  0],
                         [ 0,  0, -1,  0,  0,  1]])

        F = D_pi * D_eps @ D_a1

        return F

    def calculate_stress(self):
        eps = self.calculate_green_lagrange_strain()
        E = self.youngs_modulus

        # stress:
        sigma = eps * E + self.prestress

        return sigma

    @property
    def normal_force(self) -> float:
        biot_stress = self.calculate_stress() * self.length / self.ref_length
        A = self.area
        F = biot_stress * A

        return F

    def draw(self, item):
        sigma = self.calculate_stress()
        color = 'black'
        eta = None

        if self.calculate_stress() > 1e-3:
            color = 'blue'
            if self.tensile_strength is not None:
                sigma_max = self.tensile_strength
                eta = sigma / sigma_max
        elif self.calculate_stress() < -1e-3:
            color = 'red'
            if self.compressive_strength is not None:
                sigma_max = -self.compressive_strength
                eta = sigma / sigma_max
        elif self.tensile_strength is not None and self.compressive_strength is not None:
            eta = 0.0

        item.set_label_location(
            ref=0.5 * (self.node_a.ref_location + self.node_b.ref_location),
            act=0.5 * (self.node_a.location + self.node_b.location),
        )

        item.add_line(
            points=[
                self.node_a.ref_location,
                self.node_b.ref_location,
            ],
            layer=10,
            color='gray',
        )

        item.add_line(
            points=[
                self.node_a.location,
                self.node_b.location,
            ],
            layer=20,
            color=color,
        )

        item.add_result('Length undeformed', self.ref_length)
        item.add_result('Length', self.length)
        item.add_result('Engineering Strain', self.calculate_linear_strain())
        item.add_result('Green-Lagrange Strain', self.calculate_green_lagrange_strain())
        item.add_result('PK2 Stress', sigma)
        item.add_result('Normal Force', self.normal_force)

        if eta is not None:
            item.add_result('Degree of Utilization', eta)
